# Generic motion controller definition
# All controller plugins inherit features from this one

from __future__ import absolute_import
from __future__ import print_function

class _ControllerGeneric:
	def test(self):
		print("test supergen")

