__package_name__ = 'numpy-stl'
__import_name__ = 'stl'
__version__ = '2.7.0'
__author__ = 'Rick van Hattem'
__author_email__ = 'Wolph@Wol.ph'
__description__ = ' '.join('''
Library to make reading, writing and modifying both binary and ascii STL files
easy.
'''.split())
__url__ = 'https://github.com/WoLpH/numpy-stl/'

